{"nbformat":4,"nbformat_minor":0,"metadata":{"kernelspec":{"display_name":"Python 3","language":"python","name":"python3"},"language_info":{"codemirror_mode":{"name":"ipython","version":3},"file_extension":".py","mimetype":"text/x-python","name":"python","nbconvert_exporter":"python","pygments_lexer":"ipython3","version":"3.6.10"},"colab":{"name":"RVMReg-Solutions.ipynb","provenance":[],"collapsed_sections":[]}},"cells":[{"cell_type":"markdown","metadata":{"id":"9Qo1i1Rw39_F","colab_type":"text"},"source":["### Toy Relevance Vector Machine (Regression) Implementation"]},{"cell_type":"markdown","metadata":{"id":"Hmu6nKZo39_K","colab_type":"text"},"source":["Author: [Evgenii Egorov](mailto:egorov.evgenyy@ya.ru)"]},{"cell_type":"code","metadata":{"id":"Bk5ORP3p39_O","colab_type":"code","colab":{},"executionInfo":{"status":"ok","timestamp":1599135644117,"user_tz":-180,"elapsed":1117,"user":{"displayName":"Евгений Егоров","photoUrl":"https://lh3.googleusercontent.com/a-/AOh14GjjsFtadlJDhkK3KIjg0MoJDzyVFLbYU3bgiJhdXw=s64","userId":"14337848831696792843"}}},"source":["import numpy as np\n","import matplotlib.pyplot as plt\n","%matplotlib inline"],"execution_count":3,"outputs":[]},{"cell_type":"markdown","metadata":{"id":"Oa65cBH-39_W","colab_type":"text"},"source":["We consider the following model:\n","    \n","\\begin{equation*}\n","    \\begin{aligned}\n","    & p(t_n|x_n, w;\\beta) = \\mathcal{N}(t_n| \\textbf{w}^Tx_n, \\beta^{-1}), \\\\\n","    & p(\\textbf{t}|X,\\textbf{w};\\beta) = \\prod\\limits_{n=1}^{N}p(t_n|x_n, \\textbf{w};\\beta) = \\mathcal{N}(\\textbf{t}|X\\textbf{w}, \\beta^{-1}), \\\\\n","    & p(\\textbf{w};\\alpha) = \\prod\\limits_{d=1}^{D}\\mathcal{N}(w_d|0, \\alpha_d^{-1})=\\mathcal{N}(\\textbf{w}|0,A^{-1}).\n","    \\end{aligned}\n","\\end{equation*}  \n","\n","And for optimize the evidence we come up with the following equations:\n","\n","$$\n","\\max\\limits_{\\alpha,\\beta}\\tfrac{N}{2}\\log\\beta+\\tfrac{1}{2}\\log|A|- \\tfrac{1}{2}\\log|\\left(\\beta X^TX+A\\right)|-\\tfrac{\\beta}{2}\\|\\textbf{t}-X\\mu\\|_2^2 -\\tfrac{1}{2}\\mu^TA\\mu.\n","$$\n","\n","\\begin{equation*}\n","\\begin{aligned}\n","& \\mu^{new} = \\beta(\\beta X^T X + A)^{-1}X^T\\textbf{t},\\\\\n","& \\alpha_i^{new} = \\frac{1}{\\mu^2_i}(1-\\Sigma_{ii}^{old}\\alpha_i^{old}),\\\\\n","& \\beta^{new} = \\frac{1}{\\|t-X\\mu\\|_2^2}\\left(N-\\text{trace} (I-\\Sigma^{old}A^{old})\\right),\\\\\n","& \\Sigma^{new} = (\\beta^{new} X^TX + A^{new})^{-1}.\\\\\n","\\end{aligned}\n","\\end{equation*}\n","\n","For details and derivations please see the corresponding slides: [slides](https://drive.google.com/open?id=1ezz6LXef6pFdbljSvvZIlnmYqietQg-s)"]},{"cell_type":"markdown","metadata":{"id":"huIns4q239_Y","colab_type":"text"},"source":["##### Task 1.1\n","Please write code at the each method <i>update_%smth%</i>."]},{"cell_type":"code","metadata":{"id":"gewcIIn639_a","colab_type":"code","colab":{},"executionInfo":{"status":"ok","timestamp":1599135648227,"user_tz":-180,"elapsed":1042,"user":{"displayName":"Евгений Егоров","photoUrl":"https://lh3.googleusercontent.com/a-/AOh14GjjsFtadlJDhkK3KIjg0MoJDzyVFLbYU3bgiJhdXw=s64","userId":"14337848831696792843"}}},"source":["class RVM:\n","    def __init__(self, data):\n","        self.X = data[0]\n","        self.t = data[1]\n","        \n","        self.alpha = np.array([0.1] * self.X.shape[1])\n","        self.beta = 1.\n","        self.Sigma = None\n","        \n","        self.data_cov = None\n","        self.data_mu = None\n","        \n","        self.idx_relevant = np.array(range(self.X.shape[1]))\n","        \n","    def data_params(self):\n","        data_cov = self.X.T @ self.X\n","        data_mu = self.X.T @ self.t\n","        return data_cov, data_mu\n","    \n","    def get_sigma(self):\n","        return np.linalg.inv(self.beta * self.data_cov + np.diag(self.alpha))\n","    \n","    @staticmethod\n","    def update_mu(Sigma, beta, data_mu):\n","        return beta * Sigma @ data_mu\n","    \n","    @staticmethod\n","    def update_alpha(Sigma, beta, mu, alpha):\n","        sigma_diag = np.diag(Sigma)\n","        return (1 - sigma_diag * alpha) / mu ** 2\n","    \n","    @staticmethod\n","    def update_beta(Sigma, alpha, mu, X, t):\n","        N = X.shape[0]\n","        sigma_diag = np.diag(Sigma)\n","        return (N - np.sum(1 - sigma_diag * alpha)) / np.linalg.norm(t - X @ mu) ** 2\n","    \n","    def update_step(self):\n","        mu = self.update_mu(self.Sigma, self.beta, self.data_mu)\n","        \n","        alpha = self.update_alpha(self.Sigma, self.beta, mu, self.alpha)\n","        beta = self.update_beta(self.Sigma, self.alpha, mu, self.X, self.t)\n","        return mu, alpha, beta\n","    \n","    def elbo(self):\n","        N = self.X.shape[0]\n","        lbound = N * np.log(self.beta) + np.sum(np.log((self.alpha))) - self.beta * np.linalg.norm(self.X @ self.mu - self.t) ** 2\\\n","        - np.sum(self.mu ** 2 * self.alpha) + np.log(np.linalg.det(self.Sigma))\n","        return lbound\n","    \n","    def prune(self, threshold):\n","        idx = np.where(self.alpha < threshold)[0]\n","        if len(idx) != 0:\n","            bad = np.where(self.alpha >= threshold)[0]\n","            temp = self.idx_relevant[np.where(self.idx_relevant != -1)] \n","            temp[bad] = -1\n","            self.idx_relevant[np.where(self.idx_relevant != -1)] = temp\n","            \n","            self.X = self.X[:, idx]\n","            self.alpha = self.alpha[idx]\n","            self.mu = self.mu[idx]\n","            self.data_cov, self.data_mu = self.data_params()\n","            \n","            \n","    def fit(self, prune=1e5, max_iter=10):\n","        self.data_cov, self.data_mu = self.data_params()\n","        \n","        for i in range(max_iter):\n","            self.Sigma = self.get_sigma()\n","            self.mu, self.alpha, self.beta = self.update_step()\n","            if prune is not None:\n","                self.prune(prune)\n","            else:\n","                self.alpha = np.clip(self.alpha, 0., 1e5)\n","            print(i, self.elbo())\n","        return 0\n","    \n","    def get_relevant_features_prune(self):\n","        return self.idx_relevant[np.where(self.idx_relevant != -1)]"],"execution_count":4,"outputs":[]},{"cell_type":"markdown","metadata":{"id":"dcICjveW39_f","colab_type":"text"},"source":["Let's see how it works. We generate some sparse data."]},{"cell_type":"code","metadata":{"id":"ilHVVTtE39_g","colab_type":"code","colab":{},"executionInfo":{"status":"ok","timestamp":1599135653537,"user_tz":-180,"elapsed":1593,"user":{"displayName":"Евгений Егоров","photoUrl":"https://lh3.googleusercontent.com/a-/AOh14GjjsFtadlJDhkK3KIjg0MoJDzyVFLbYU3bgiJhdXw=s64","userId":"14337848831696792843"}}},"source":["from sklearn.datasets import make_regression\n","from sklearn.metrics import r2_score"],"execution_count":5,"outputs":[]},{"cell_type":"code","metadata":{"id":"ODTx2hQ_39_l","colab_type":"code","colab":{"base_uri":"https://localhost:8080/","height":329},"executionInfo":{"status":"ok","timestamp":1599135656693,"user_tz":-180,"elapsed":920,"user":{"displayName":"Евгений Егоров","photoUrl":"https://lh3.googleusercontent.com/a-/AOh14GjjsFtadlJDhkK3KIjg0MoJDzyVFLbYU3bgiJhdXw=s64","userId":"14337848831696792843"}},"outputId":"6c0c9273-cbb9-4653-e235-a107625bb862"},"source":["X, t, coef_true = make_regression(100, 100, 3, coef=True)\n","print(coef_true)\n","print(np.where(coef_true != 0)[0])"],"execution_count":6,"outputs":[{"output_type":"stream","text":["[ 0.          0.          0.          0.          0.          0.\n","  0.          0.          0.          0.          0.          0.\n","  0.          0.          0.          0.          0.          0.\n","  0.          0.          0.          0.          0.          0.\n","  0.          0.          0.          0.          0.          0.\n"," 83.8110758   0.          0.          0.          0.          0.\n","  0.          0.          0.          0.          0.          0.\n","  0.          0.          0.          0.          0.          0.\n","  0.          0.          0.          0.          0.          0.\n","  0.          0.          0.          0.          0.          0.\n","  0.          0.          0.          0.          0.          0.\n"," 63.64974908  0.          0.          0.          0.          0.\n","  0.          0.          0.          0.          0.          0.\n","  0.          0.          0.          0.          0.          0.\n","  0.          0.          0.          0.          0.          0.\n","  0.          0.          0.          0.         81.40968181  0.\n","  0.          0.          0.          0.        ]\n","[30 66 94]\n"],"name":"stdout"}]},{"cell_type":"markdown","metadata":{"id":"SBDZmLJr39_r","colab_type":"text"},"source":["And now we expect that RVM will return for us exactly these non-zero features.m"]},{"cell_type":"code","metadata":{"id":"d28f3f0N39_s","colab_type":"code","colab":{},"executionInfo":{"status":"ok","timestamp":1599135664459,"user_tz":-180,"elapsed":847,"user":{"displayName":"Евгений Егоров","photoUrl":"https://lh3.googleusercontent.com/a-/AOh14GjjsFtadlJDhkK3KIjg0MoJDzyVFLbYU3bgiJhdXw=s64","userId":"14337848831696792843"}}},"source":["reg = RVM((X, t))"],"execution_count":7,"outputs":[]},{"cell_type":"code","metadata":{"id":"76lp80f239_w","colab_type":"code","colab":{"base_uri":"https://localhost:8080/","height":208},"executionInfo":{"status":"ok","timestamp":1599135667284,"user_tz":-180,"elapsed":971,"user":{"displayName":"Евгений Егоров","photoUrl":"https://lh3.googleusercontent.com/a-/AOh14GjjsFtadlJDhkK3KIjg0MoJDzyVFLbYU3bgiJhdXw=s64","userId":"14337848831696792843"}},"outputId":"90c8e08c-e034-4e0d-8498-dbfe9d682336"},"source":["reg.fit(1e5)"],"execution_count":8,"outputs":[{"output_type":"stream","text":["0 -571.4852563595872\n","1 503.0053096111681\n","2 3668.457456244226\n","3 5895.977785780902\n","4 5817.316452570509\n","5 5785.002094980026\n","6 5775.732725762403\n","7 5871.9997629475765\n","8 5918.5623620195865\n","9 5814.520528609179\n"],"name":"stdout"},{"output_type":"execute_result","data":{"text/plain":["0"]},"metadata":{"tags":[]},"execution_count":8}]},{"cell_type":"code","metadata":{"id":"IbX4eUe639_1","colab_type":"code","colab":{"base_uri":"https://localhost:8080/","height":34},"executionInfo":{"status":"ok","timestamp":1599135671430,"user_tz":-180,"elapsed":863,"user":{"displayName":"Евгений Егоров","photoUrl":"https://lh3.googleusercontent.com/a-/AOh14GjjsFtadlJDhkK3KIjg0MoJDzyVFLbYU3bgiJhdXw=s64","userId":"14337848831696792843"}},"outputId":"7c8e1b0d-29d8-4861-99d1-2f6d9f867646"},"source":["reg.get_relevant_features_prune()"],"execution_count":9,"outputs":[{"output_type":"execute_result","data":{"text/plain":["array([30, 66, 94])"]},"metadata":{"tags":[]},"execution_count":9}]},{"cell_type":"markdown","metadata":{"id":"NKud8R6R39_6","colab_type":"text"},"source":["Eeee!"]}]}